/*
   program : recon

   $Source: /home/yee/C/RCS/recon.c,v $
   $Author: yee $
   $Date: 1999/03/09 01:52:38 $
   $Revision: 1.40 $

   This is the driver for the functions in backproject.c

   Compile Options :
     1) #define READ_DATA_BY_LINE to read one projection at a time.

   Parameters : see the documentation for ParseArg()

   Assumptions : none known

   Bugs : none known

   To Do :
     1) support multiple sizes of float.

   $Log: recon.c,v $
   Revision 1.40  1999/03/09 01:52:38  yee
   *** empty log message ***

   Revision 1.39  1999/02/10 16:39:47  yee
   *** empty log message ***

   Revision 1.38  1999/01/24 06:34:02  yee
   bugfix setting square pixels.

   Revision 1.30  1998/08/29 16:04:36  yee
   changed options to allow primary definition of fov rather than pixel size.

   Revision 1.29  1998/08/27 22:56:56  yee
   working for colinear fan and parallel geometries.

   Copyright 1997,1998,1999 by Conway Yee.  All rights reserved.
 */

char recon[] = "$Id: recon.c,v 1.40 1999/03/09 01:52:38 yee Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "style.h"
#include "array.h"
#include "filter.h"
#include "mutil.h"
#include "util.h"

#include "backproject.h"
#include "ImageHeader.h"

PUBLIC char *optarg;               /* used by getopt() */
PUBLIC int optind=0;
PUBLIC int opterr=0;
PUBLIC int optopt=0;

BOOLEAN verbose=FALSE;

PRIVATE ImageHeader *ParseArg(ImageHeader *,int,char **,FILE **,FILE **, double (**)());
 
/*****************************************************************************/
/*****************************************************************************/

/* this the main module driver for filter backprojection. */

PUBLIC int main(int argc,char **argv)

{
  ImageHeader *ProjHeader;         /* header of projection data */
  ImageHeader *ReconHeader;        /* header of reconstructed image */
  FILE *inptr=stdin;               /* input file pointer */
  FILE *outptr=stdout;             /* output file pointer */
  double (*window)()=NULL;         /* windowing filter to be used */

  AllocateImageHeader(ReconHeader,NDIM);   /* setup reconstructed image data */
  ReconHeader=ParseArg(ReconHeader,argc,argv,&inptr,&outptr,&window);
  ReconHeader->fptr=outptr;

  ReconHeader->data.f = (RECON_FLOAT **)mdmalloc
        (NDIM,sizeof(RECON_FLOAT),ALLOC_DISCRETE,ReconHeader->lim);
  if(ReconHeader->data.f<0) {
    perror("recon : mdmalloc failure on reconstructed image array ");
    return(errno);
  }
  AllocateImageHeader(ProjHeader,NDIM);             /* setup projection data */

  if(!ioImageHeader(ProjHeader,fread,inptr)) {
    errno=HEADER_ERROR;
    perror("recon : unable to read projection header ");
    return(errno);
  }
  ProjHeader->fptr=inptr;

#ifndef READ_DATA_BY_LINE
  ProjHeader->data.c = (RECON_COMPLEX **)mdmalloc
       (NDIM,sizeof(RECON_COMPLEX),ALLOC_CONTIGUOUS,ProjHeader->lim);
  if(ProjHeader->data.c<0) {
    perror("recon : mdmalloc failure on projection data array ");
    return(errno);
  }
#endif READ_DATA_BY_LINE
                              /* transfer personal data from projection
                                 data to reconstructed image */
  ImageHeaderClonePersonalData(ProjHeader,ReconHeader);

  if(ProjHeader->fsize == sizeof(RECON_FLOAT)) {     /* read projection data */

#ifndef READ_DATA_BY_LINE
    mdio(inptr,2,ProjHeader->fsize,sizeof(RECON_COMPLEX),ProjHeader->data.c,
         fread,ProjHeader->lim);
#endif !READ_DATA_BY_LINE
  }  else {
    fprintf(stderr,"recon : incompatible float size\n");
    fprintf(stderr,"        file : %1d bytes      program : %1d bytes\n",
            ProjHeader->fsize,sizeof(RECON_FLOAT));
    errno = INCOMPATIBLE_DATA;
    return(errno);
  }
  FilteredBackproject(ReconHeader,ProjHeader,window);   /* reconstruct image */

  if(!ioImageHeader(ReconHeader,fwrite,outptr)) {            /* output image */
    errno=HEADER_ERROR;
    perror("recon : unable to write output header ");
    return(errno);
  }
  mdio(outptr,2,sizeof(RECON_FLOAT),sizeof(RECON_FLOAT),ReconHeader->data.f,
       fwrite,ReconHeader->lim);

  FreeImageHeader(ProjHeader);                                   /* clean up */
  FreeImageHeader(ReconHeader);
  fclose(inptr);
  fclose(outptr);

  return(errno);
}

/*****************************************************************************/

/* This function processes the command line arguments of recon.
   See the declaration of *errormsg[] for legal options. */

PRIVATE ImageHeader *ParseArg
(
 ImageHeader *h,                    /* image header */
 int argc,                          /* argument count */
 char **argv,                       /* arguments */
 FILE **inptr,                      /* input file pointer */
 FILE **outptr,                     /* output file pointer */
 double (**window)()                /* Windowing filter to be used */
)
{ /* ParseArg() */
  char opt;                                                        /* option */
  char *infname=NULL;                              /* filename of input file */
  char *outfname=NULL;                            /* filename of output file */
  BOOLEAN error=FALSE;                        /* tests if error in arguments */
  double *range=calloc(1,sizeof(double)*NDIM);              /* field of view */

  char ReconOptions[]="i:o:w:X:Y:x:y:v";
  char *errormsg[]={
    "recon [-i <name>] [-o <name>] [-X <float>] [-Y <float>] [-x <int>] [-y <int>]",
    "   -i <input file name>",
    "   -o <output file name>",
    "   -w <window function>",
    "   -X <range in X>",
    "   -Y <range in Y>",
    "   -x <number of pixels in X>",
    "   -y <number of pixels in Y>",
    "   -v verbose mode",
    NULL
  };

  do {
    opt = getopt(argc,argv,ReconOptions);
    switch(opt) {
    case 'i':                                              /* input filename */
      *inptr=fopen(optarg,"r");
      if(*inptr == NULL) {
        perror("recon : unable to open input file");
        error=TRUE;
      } else {
        infname=malloc(strlen(optarg)+1);
        strcpy(infname,optarg);
      }
      break;
    case 'o':                                             /* output filename */
      *outptr=fopen(optarg,"w+");
      if(*outptr == NULL) {
        perror("recon : unable to open output file");
        error=TRUE;
      } else {
        outfname=malloc(strlen(optarg)+1);
        strcpy(outfname,optarg);
      }
      break;
    case 'w':                                          /* window function */
      STOLOWER(optarg);
      fprintf(stderr,"recon : %s filter not supported\n",optarg);
       *window=NULL;
      break;
    case 'X':                                             /* range in X */
      range[X] = atof(optarg);
      if ((errno == ERANGE)||(range[X]<0)) {
        perror("recon : illegal distance between samples ");
        error=TRUE;
      }
      break;
    case 'Y':                                             /* range in Y */
      range[Y] = atof(optarg);
      if ((errno == ERANGE)||(range[Y]<0)) {
        perror("recon : illegal distance between samples ");
        error=TRUE;
      }
      break;
    case 'x':                                       /* number of pixels in X */
      {
        int num;               /* number read as option */
        h->dim[X] = num = atoi(optarg);
        h->lim[LOX] = num/2 - num+1;
        h->lim[HIX] = num/2;
        if ((errno == ERANGE)||(h->ndim<0)) {
          perror("recon : illegal # of pixels in X ");
          error=TRUE;
        }
      }
      break;
    case 'y':                                       /* number of pixels in X */
      {
        int num;               /* number read as option */
        h->dim[Y] = num = atoi(optarg);
        h->lim[LOY] = num/2 - num+1;
        h->lim[HIY] = num/2;
        if ((errno == ERANGE)||(h->lim[HIY]<0)) {
          perror("recon : illegal # of pixels in Y ");
          error=TRUE;
        }
      }
      break;
    case 'v':                                                /* verbose mode */
      verbose=TRUE;
      break;
    case ':':                                           /* missing parameter */
      perror("recon : missing parameter in option");
      error=TRUE;
      break;
    case '?':                                              /* illegal option */
      perror("recon : illegal option");
      error=TRUE;
      break;
    case -1:                                           /* all options parsed */
      if((h->dim[Y]<=0)||(h->dim[X]<=0)) {   /* number of pixels */
        h->lim[HIX]=h->lim[HIY]=MAX(h->lim[HIX],h->lim[HIY]);
        h->lim[LOX]=h->lim[LOY]=MIN(h->lim[LOX],h->lim[LOY]);
        h->dim[X]=h->dim[Y]=MAX(h->dim[X],h->dim[Y]);
        if(h->lim[HIX]>0)
          fprintf(stderr,"recon : setting square image : %1d\n",h->lim[HIX]);
        else {
          fprintf(stderr,"recon : pixel number not set\n");
          error=TRUE;
        }
      }  /* number of pixels */

      { /* set pixel size */
        int i; /* counter */
        for(i=0;i<NDIM;i++) h->tau[i] = range[i]/h->dim[i];
      } /* set pixel size */

      if((h->tau[X]<=0.0)||(h->tau[Y]<=0.0)) {  /* pixel size */
        h->tau[X]=h->tau[Y]=MAX(h->tau[X],h->tau[Y]);
        if(h->tau[X]>0)
          fprintf(stderr,"recon : setting square pixel : %g\n",h->tau[X]);
        else {
          fprintf(stderr,"recon : pixel size not set\n");
          error=TRUE;
        }
      }  /* pixel size */
       
      if(error) {
        USAGE(errormsg)
      } else if(verbose) {
        fprintf(stderr,"recon : input filename : %s\n",infname);
        fprintf(stderr,"recon : output filename : %s\n",outfname);
        fprintf(stderr,"recon : %f distance between pixels in X\n",h->tau[X]);
        fprintf(stderr,"recon : %f distance between pixels in Y\n",h->tau[Y]);
        fprintf(stderr,"recon : %d pixels in X\n",h->dim[X]);
        fprintf(stderr,"recon : %d pixels in Y\n",h->dim[Y]);
      }

      free(infname);
      free(outfname);

      break;
    default :
      errno = ILLEGAL_ARGUMENTS;
      error=TRUE;
    }
  } while(opt != -1);

  free(range);
  return(h);

} /* ParseArg() */
